r"""Tests for the inox.lift module."""

from functools import partial

from inox.lift import automask, inner, outer
from inox.tree import Mask, Static


def test_automask():
    def is_static(x):
        assert type(x) is Static
        return x

    def is_masked(x):
        assert type(x) is Mask
        return x

    def is_not_masked(x):
        assert type(x) is not Mask
        return x

    # inner & outer
    assert is_masked(inner(is_not_masked)("leaf"))
    assert is_masked(inner(is_not_masked)(Mask("leaf")))
    assert is_static(inner(is_static)(Static("leaf")))
    assert is_not_masked(outer(is_masked)("leaf"))
    assert is_not_masked(outer(is_masked)(Mask("leaf")))
    assert is_static(outer(is_static)(Static("leaf")))

    def f(x):
        return x

    assert outer(inner(f)) is f
    assert inner(outer(f)) is f
    assert inner(f) is inner(f)

    # automask
    assert is_not_masked(automask(partial)(is_not_masked)("leaf"))
    assert is_not_masked(automask(partial)(is_not_masked)(Mask("leaf")))
    assert is_static(automask(partial)(is_static)(Static("leaf")))

    assert automask(lambda f: f)(f) is f
    assert automask(partial)(f) is not f



r"""Extended utilities for tree-like data structures."""

__all__ = [
    "Namespace",
    "Partial",
    "Static",
    "mask",
    "unmask",
    "partition",
    "combine",
    "prepr",
]

import dataclasses
import jax._src.tree_util as jtu
import numpy as np

from jax import Array
from textwrap import indent
from typing import Any, Callable, Dict, Hashable, List, Tuple, TypeVar, Union
from warnings import warn

PyTree = TypeVar("PyTree", bound=Any)
PyTreeDef = TypeVar("PyTreeDef")


def is_array(x: Any) -> bool:
    return isinstance(x, np.ndarray) or isinstance(x, Array)


def is_atom(x: Any) -> bool:
    return x is None or jtu.all_leaves((x,))


class PyTreeMeta(type):
    r"""PyTree meta-class."""

    def __new__(cls, *args, **kwargs) -> type:
        cls = super().__new__(cls, *args, **kwargs)

        if hasattr(cls, "tree_flatten_with_keys"):
            jtu.register_pytree_with_keys_class(cls)
        else:
            jtu.register_pytree_node_class(cls)

        return cls


class Namespace(metaclass=PyTreeMeta):
    r"""PyTree class for name-value mappings.

    Arguments:
        kwargs: A name-value mapping.

    Example:
        >>> tree = Namespace(a=1, b='2'); tree
        Namespace(
          a = 1,
          b = '2'
        )
        >>> tree.c = [3, False]; tree
        Namespace(
          a = 1,
          b = '2',
          c = [3, False]
        )
        >>> jax.tree.leaves(tree)
        [1, '2', 3, False]
    """

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self) -> str:
        return prepr(self)

    def tree_repr(self, **kwargs) -> str:
        private = kwargs.get("private", False)

        lines = (
            f"{name} = {prepr(getattr(self, name), **kwargs)}"
            for name in sorted(self.__dict__.keys())
            if not name.startswith("_") or private
        )

        lines = ",\n".join(lines)

        if lines:
            lines = "\n" + indent(lines, "  ") + "\n"

        return f"{self.__class__.__name__}({lines})"

    def tree_flatten(self):
        if self.__dict__:
            names, values = zip(*sorted(self.__dict__.items()))
        else:
            names, values = (), ()

        return values, names

    def tree_flatten_with_keys(self):
        values, names = self.tree_flatten()
        keys = map(jtu.GetAttrKey, names)

        return list(zip(keys, values)), names

    @classmethod
    def tree_unflatten(cls, names, values):
        self = object.__new__(cls)
        self.__dict__ = dict(zip(names, values))

        return self


class Partial(Namespace):
    r"""A version of :class:`functools.partial` that is a PyTree.

    Arguments:
        func: A function.
        args: Positional arguments for future calls.
        kwds: Keyword arguments for future calls.

    Examples:
        >>> increment = Partial(jax.numpy.add, 1)
        >>> increment(2)
        Array(3, dtype=int32, weak_type=True)

        >>> println = Partial(print, sep='\n')
        >>> println('Hello', 'World!')
        Hello
        World!
    """

    def __init__(self, func: Callable, *args: Any, **kwds: Any):
        self.func = func
        self.args = args
        self.kwds = kwds

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.func(*self.args, *args, **self.kwds, **kwds)


class Static(metaclass=PyTreeMeta):
    r"""Wraps an hashable value as a leafless PyTree.

    Arguments:
        value: An hashable value to wrap.

    Example:
        >>> tree = Static((0, 'one', None))
        >>> tree.value
        (0, 'one', None)
        >>> jax.tree.leaves(tree)
        []
        >>> jax.tree.structure(tree)
        PyTreeDef(CustomNode(Static[(0, 'one', None)], []))
    """

    def __init__(self, value: Hashable):
        if not isinstance(value, Hashable):
            warn(f"considering a non-hashable object ('{type(value).__name__}') static could lead to frequent JIT recompilations.", stacklevel=2)  # fmt: off
        elif callable(value):
            if "<lambda>" in value.__qualname__ or "<locals>" in value.__qualname__:
                warn(f"considering a local function ('{value.__qualname__}') static could lead to frequent JIT recompilations.", stacklevel=2)  # fmt: off

        self.value = value

    def __eq__(self, other: Any) -> bool:
        return type(self) is type(other) and self.value == other.value

    def __hash__(self) -> int:
        return hash((type(self), self.value))

    def __repr__(self) -> str:
        return self.tree_repr()

    def tree_repr(self, **kwargs) -> str:
        return f"{self.__class__.__name__}({prepr(self.value, **kwargs)})"

    def tree_flatten(self):
        return (), self.value

    def tree_flatten_with_keys(self):
        return self.tree_flatten()

    @classmethod
    def tree_unflatten(cls, value, _):
        self = object.__new__(cls)
        self.value = value

        return self


class Mask(Static):
    pass


def mask(
    tree: PyTree,
    is_static: Callable[[Any], bool] = None,
) -> PyTree:
    r"""Masks the static leaves of a tree.

    The structure of the tree remains unchanged, but leaves that are considered static
    are masked, which hides them from :func:`jax.tree.leaves` and :func:`jax.tree.map`.
    Applying :func:`inox.tree.mask` more than once leads to the same tree.

    See also:
        :func:`inox.tree.unmask`

    Arguments:
        tree: The tree to mask.
        is_static: A predicate for what to consider static. If :py:`None`,
            all non-array leaves are considered static.

    Returns:
        The masked tree.

    Example:
        >>> tree = [1, jax.numpy.arange(2), 'three']
        >>> jax.tree.leaves(tree)
        [1, Array([0, 1], dtype=int32), 'three']
        >>> tree = mask(tree); tree
        [Mask(1), Array([0, 1], dtype=int32), Mask('three')]
        >>> jax.tree.leaves(tree)
        [Array([0, 1], dtype=int32)]
        >>> unmask(tree)
        [1, Array([0, 1], dtype=int32), 'three']
    """

    if is_static is None:
        is_static = lambda x: not is_array(x)

    return jtu.tree_map(
        f=lambda x: Mask(x) if is_static(x) else x,
        tree=tree,
    )


def unmask(tree: PyTree) -> PyTree:
    r"""Unmasks the masked leaves of a tree.

    See also:
        :func:`inox.tree.mask`

    Arguments:
        tree: The masked tree to unmask.

    Returns:
        The unmasked tree.
    """

    is_masked = lambda x: type(x) is Mask

    return jtu.tree_map(
        f=lambda x: x.value if is_masked(x) else x,
        tree=tree,
        is_leaf=is_masked,
    )


def partition(
    tree: PyTree,
    *filters: Union[type, Callable[[Any], bool]],
    is_leaf: Callable[[Any], bool] = None,
) -> Tuple[PyTreeDef, Dict[str, Any]]:
    r"""Flattens a tree and partitions its leaves.

    The leaves are partitioned into a set of path-leaf mappings. The mapping in which a
    leaf is contained is chosen according to its oldest (closest to the root) ancestor
    that satisfies a constraint. If a node satisfies several constraints, the first one
    is selected. The last mapping is dedicated to leaves that do not satisfy any
    constraint.

    See also:
        :func:`inox.tree.combine`

    Arguments:
        tree: The tree to flatten.
        filters: A set of filtering constraints. Types are transformed into
            :py:`isinstance` constraints.
        is_leaf: A predicate for what to consider as a leaf.

    Returns:
        The tree definition and leaf partitions.

    Example:
        >>> tree = Namespace(a=1, b=jax.numpy.arange(2), c=['three', False])
        >>> treedef, leaves = inox.tree.partition(tree)
        >>> leaves
        {'.a': 1, '.b': Array([0, 1], dtype=int32), '.c[0]': 'three', '.c[1]': False}
        >>> treedef, arrays, others = inox.tree.partition(tree, jax.Array)
        >>> arrays
        {'.b': Array([0, 1], dtype=int32)}
        >>> others
        {'.a': 1, '.c[0]': 'three', '.c[1]': False}
    """

    treedef = jtu.tree_structure(tree, is_leaf)
    leaves = [{} for _ in filters] + [{}]

    def factory(filtr):
        if isinstance(filtr, type):
            return lambda x: isinstance(x, filtr)
        else:
            return filtr

    filters = list(map(factory, filters))

    if is_leaf is None:
        is_node = lambda x: any(filtr(x) for filtr in filters)
    else:
        is_node = lambda x: any(filtr(x) for filtr in filters) or is_leaf(x)

    def f(node_path, node):
        for i, filtr in enumerate(filters):  # noqa: B007
            if filtr(node):
                break
        else:
            i = len(filters)

        def g(leaf_path, leaf):
            key = jtu.keystr(node_path + leaf_path)
            leaves[i][key] = leaf

        jtu.tree_map_with_path(f=g, tree=node, is_leaf=is_leaf)

    jtu.tree_map_with_path(f=f, tree=tree, is_leaf=is_node)

    return treedef, *leaves


def combine(
    treedef: PyTreeDef,
    *leaves: Dict[str, Any],
) -> PyTree:
    r"""Reconstructs a tree from the tree definition and leaf partitions.

    See also:
        :func:`inox.tree.partition`

    Arguments:
        treedef: The tree definition.
        leaves: The set of leaf partitions.

    Returns:
        The reconstructed tree.

    Example:
        >>> tree = Namespace(a=1, b=jax.numpy.arange(2), c=['three', False])
        >>> treedef, arrays, others = inox.tree.partition(tree, jax.Array)
        >>> others = {key: str(leaf).upper() for key, leaf in others.items()}
        >>> inox.tree.combine(treedef, arrays, others)
        Namespace(
          a = '1',
          b = int32[2],
          c = ['THREE', 'FALSE']
        )
    """

    missing = []
    leaves = {key: leaf for partition in leaves for key, leaf in partition.items()}

    def f(path, leaf):
        key = jtu.keystr(path)

        if key in leaves:
            leaf = leaves.pop(key)
        else:
            missing.append(key)

        return leaf

    tree = jtu.tree_unflatten(treedef, [object()] * treedef.num_leaves)
    tree = jtu.tree_map_with_path(f, tree)

    if missing:
        keys = ", ".join(f'"{key}"' for key in missing)

        raise KeyError(f"Missing key(s) in leaves: {keys}.")

    if leaves:
        keys = ", ".join(f'"{key}"' for key in leaves)

        raise KeyError(f"Unexpected key(s) in leaves: {keys}.")

    return tree


def find_duplicates(tree: PyTree) -> List[List[str]]:
    r"""Finds duplicated nodes in a tree.

    This function traverses the tree and collects paths to nodes that have the same
    identity. It returns a list of groups, where each group is a list of paths to nodes
    that are duplicates.

    Arguments:
        tree: A tree.

    Returns:
        The list of duplicated path groups.

    Example:
        >>> a = []
        >>> b = {"c": a, "d": None, "e": ["f", a]}
        >>> b["h"] = b
        >>> find_duplicates(b)
        [['', "['h']"], ["['c']", "['e'][1]"]]
    """

    nodes = {}

    def f(path, node):
        if is_atom(node):
            return

        group = nodes.setdefault(id(node), [])
        group.append(jtu.keystr(path))

        if len(group) > 1:
            return

        children, _ = jtu.flatten_one_level_with_keys(node)

        for key, child in children:
            f((*path, key), child)

    f((), tree)

    duplicates = [group for group in nodes.values() if len(group) > 1]

    return duplicates


def prepr(
    tree: PyTree,
    linewidth: int = 88,
    typeonly: bool = True,
    **kwargs,
) -> str:
    r"""Creates a pretty representation of a tree.

    Arguments:
        tree: The tree to represent.
        linewidth: The maximum line width before elements of tuples, lists and dicts
            are represented on separate lines.
        typeonly: Whether to represent the type of arrays instead of their elements.

    Returns:
        The representation string.

    Example:
        >>> tree = [1, 'two', (True, False), list(range(5)), {'6': jax.numpy.arange(7)}]
        >>> tree.append(tree)
        >>> print(inox.tree.prepr(tree))
        [1, 'two', (True, False), [0, 1, 2, 3, 4], {'6': int32[7]}, [...]]
    """

    kwargs.update(
        linewidth=linewidth,
        typeonly=typeonly,
    )

    ancestors = kwargs.setdefault("ancestors", set())

    if id(tree) in ancestors:
        if hasattr(tree, "tree_repr"):
            return f"{type(tree).__name__}(...)"
        elif dataclasses._is_dataclass_instance(tree):
            return f"{type(tree).__name__}(...)"
        elif isinstance(tree, tuple):
            if hasattr(tree, "_fields"):
                return f"{type(tree).__name__}(...)"
            else:
                return "(...)"
        elif isinstance(tree, list):
            return "[...]"
        elif isinstance(tree, dict):
            return "{...}"

    ancestors.add(id(tree))

    if hasattr(tree, "tree_repr"):
        return tree.tree_repr(**kwargs)
    elif dataclasses._is_dataclass_instance(tree):
        bra, ket = f"{type(tree).__name__}(", ")"
        lines = [
            f"{field.name}={prepr(getattr(tree, field.name), **kwargs)}"
            for field in dataclasses.fields(tree)
            if field.repr
        ]
    elif isinstance(tree, tuple):
        if hasattr(tree, "_fields"):
            bra, ket = f"{type(tree).__name__}(", ")"
            lines = [
                f"{field}={prepr(value, **kwargs)}" for field, value in tree._asdict().items()
            ]
        else:
            bra, ket = "(", ")"
            lines = [prepr(x, **kwargs) for x in tree]
    elif isinstance(tree, list):
        bra, ket = "[", "]"
        lines = [prepr(x, **kwargs) for x in tree]
    elif isinstance(tree, dict):
        bra, ket = "{", "}"
        lines = [f"{repr(key)}: {prepr(value, **kwargs)}" for key, value in tree.items()]
    elif is_array(tree) and typeonly:
        return f"{tree.dtype}{list(tree.shape)}"
    else:
        return repr(tree).strip(" \n")

    ancestors.remove(id(tree))

    if any("\n" in line for line in lines):
        lines = ",\n".join(lines)
    elif sum(len(line) + 2 for line in lines) > linewidth:
        lines = ",\n".join(lines)
    else:
        lines = ", ".join(lines)

    if "\n" in lines:
        lines = "\n" + indent(lines, "  ") + "\n"

    return f"{bra}{lines}{ket}"



r"""Sharing modules

In JAX, implicit shared references (objects with the same Python :py:`id`) to a node in
a tree are treated as distinct objects. While this design choice is reasonable, it makes
it difficult to express that two layers in a module are identical, and preserve their
shared identity through transformations.

Inox provides a mechanism to express explicitly such identity. During flattening of a
:class:`Scope` module, if several :class:`Reference` instances in the tree share the
same identification tag, the first occurence (depth-first order) is preserved, while the
following occurences are pruned. During unflattening, the pruned occurences are filled
in with the preserved occurence, which preserves their shared identity.

.. code-block:: python

    import inox
    import inox.nn as nn
    import jax
    import jax.numpy as jnp

    class WeightSharingMLP(nn.Scope):
        def __init__(self, key):
            keys = jax.random.split(key, 3)

            self.l1 = nn.Linear(in_features=64, out_features=64, key=keys[0])
            self.l3 = nn.Linear(in_features=64, out_features=64, key=keys[1])
            self.l4 = nn.Linear(in_features=64, out_features=64, key=keys[2])

            self.l1 = nn.Reference('l1', self.l1)
            self.l2 = self.l1  # tied layer
            self.l3.weight = nn.Reference('l3.weight', self.l3.weight)
            self.l4.weight = self.l3.weight  # tied parameter

            self.relu = nn.ReLU()

        def __call__(self, x):  # standard __call__
            x = self.l1(x)
            x = self.l2(self.relu(x))
            x = self.l3(self.relu(x))
            x = self.l4(self.relu(x))

            return x

    key = jax.random.key(0)
    model = WeightSharingMLP(key)
    static, params, others = model.partition(nn.Parameter)

    print(inox.tree.prepr(params))  # does not contain 'l2' and 'l4.weight'

.. code-block:: text

    {
      '.l1.value.bias.value': float32[64],
      '.l1.value.weight.value': float32[64, 64],
      '.l3.bias.value': float32[64],
      '.l3.weight.value.value': float32[64, 64],
      '.l4.bias.value': float32[64]
    }
"""

__all__ = [
    "Scope",
    "Reference",
]

import jax.tree_util as jtu

from typing import Any, Hashable, Iterator

import inox.nn as nn
import inox.tree


class Scope(nn.Module):
    r"""Subclass of :class:`inox.nn.module.Module` which handles shared object
    references within its scope.

    Arguments:
        kwargs: A name-value mapping.
    """

    def tree_flatten(self):
        children, meta = super().tree_flatten()
        memory = {}

        def prune(tree):
            def f(x):
                if isinstance(x, Reference):
                    if x.tag in memory:
                        y = memory[x.tag]
                    else:
                        y = memory[x.tag] = Reference(x.tag, None)

                    new = Reference(x.tag, None)

                    if x.obj is None:
                        pass
                    elif y.obj is None:
                        y.obj = x.obj
                        new.obj = prune(x.obj)
                    else:
                        assert x.obj is y.obj, f"#{x.tag} has conflicting identities"

                    return new

                return x

            def is_leaf(x):
                return isinstance(x, Reference) or isinstance(x, Scope)

            return jtu.tree_map(f, tree, is_leaf=is_leaf)

        return prune(children), meta

    @classmethod
    def tree_unflatten(cls, meta, children):
        memory = {}

        def unprune(tree):
            def f(x):
                if isinstance(x, Reference):
                    if x.tag in memory:
                        y, new = memory[x.tag]
                    else:
                        y, new = memory[x.tag] = Reference(x.tag, None), Reference(x.tag, None)

                    if x.obj is None:
                        pass
                    elif y.obj is None:
                        y.obj = x.obj
                        new.obj = unprune(x.obj)
                    else:
                        assert x.obj is y.obj, f"#{x.tag} has conflicting identities"

                    return new

                return x

            def is_leaf(x):
                return isinstance(x, Reference) or isinstance(x, Scope)

            return jtu.tree_map(f, tree, is_leaf=is_leaf)

        return super().tree_unflatten(meta, unprune(children))


class Reference(metaclass=inox.tree.PyTreeMeta):
    r"""Creates a reference to an object.

    A :class:`Reference` instance forwards :py:`__call__`, :py:`__iter__`,
    :py:`__getattr__`, and :py:`__getitem__` operations to the object it references. For
    arithmetic operations (`+`, `*`, ...), use :py:`ref.obj` directly instead.

    Arguments:
        tag: An identification tag.
        obj: The object to reference.

    Example:
        >>> dummy = ["zero", nn.Parameter(jax.numpy.ones((2, 3)))]
        >>> dummy = Reference("dummy-list", dummy)
        >>> dummy  # repr preceded by asterisk
        *['zero', Parameter(float32[2, 3])]
        >>> len(dummy)
        2
        >>> "zero" in dummy
        True
        >>> dummy[1].shape
        (2, 3)
    """

    def __init__(self, tag: Hashable, obj: Any):
        self.tag = tag
        self.obj = obj

    def __call__(self, *args, **kwargs) -> Any:
        return self.obj(*args, **kwargs)

    def __delattr__(self, name: str):
        if name in ("tag", "obj"):
            object.__delattr__(self, name)
        else:
            delattr(self.obj, name)

    def __getattr__(self, name: str) -> Any:
        if name in ("tag", "obj"):
            return object.__getattr__(self, name)
        else:
            return getattr(self.obj, name)

    def __setattr__(self, name: str, value: Any):
        if name in ("tag", "obj"):
            object.__setattr__(self, name, value)
        else:
            setattr(self.obj, name, value)

    def __delitem__(self, key: Hashable):
        del self.obj[key]

    def __getitem__(self, key: Hashable) -> Any:
        return self.obj[key]

    def __setitem__(self, key: Hashable, value: Any):
        self.obj[key] = value

    def __contains__(self, key: Hashable) -> bool:
        return key in self.obj

    def __len__(self) -> int:
        return len(self.obj)

    def __iter__(self) -> Iterator[Any]:
        return iter(self.obj)

    def __repr__(self) -> str:
        return self.tree_repr()

    def tree_repr(self, **kwargs) -> str:
        if self.obj is None:
            return f"#{self.tag}"
        else:
            return f"*{inox.tree.prepr(self.obj, **kwargs)}"

    def tree_flatten(self):
        return [self.obj], self.tag

    def tree_flatten_with_keys(self):
        return [(jtu.GetAttrKey("obj"), self.obj)], self.tag

    @classmethod
    def tree_unflatten(cls, tag, children):
        self = object.__new__(cls)
        self.tag = tag
        self.obj = next(iter(children))

        return self



r"""Attention layers"""

__all__ = [
    "MultiheadAttention",
]

import jax
import jax.numpy as jnp

from einops import rearrange
from jax import Array
from typing import Union

from .linear import Linear
from .module import Module
from ..random import get_rng


class MultiheadAttention(Module):
    r"""Creates a multihead attention layer.

    .. math:: Y = \sum_i
        \mathrm{attention}(X_q W_q^i + b_q^i, X_k W_k^i + b_k^i, X_v W_v^i + b_v^i) W_y^i

    where

    .. math:: \mathrm{attention}(Q, K, V) =
        \mathrm{softmax}\left( \frac{Q K^T}{\sqrt{H}} \right) V

    denotes the scaled dot-product attention.

    References:
        | Attention Is All You Need (Vaswani et al., 2023)
        | https://arxiv.org/abs/1706.03762

    Arguments:
        in_features: The number of input features :math:`C`.
        out_features: The number of output features :math:`C'`.
            If :py:`None`, :math:`C' = C`.
        hid_features: The number of hidden features :math:`H` per head.
            If :py:`None`, :math:`H = \frac{C}{N}`.
        heads: The number of attention heads :math:`N`.
        bias: Whether the layer learns additive biases :math:`(b_q, b_k, b_v)` or not.
        causal: Whether the attention mask is causal or not. If :py:`True`, the
            :math:`i`-th query is only allowed to attend the :math:`j`-th key if
            :math:`j - i \leq T - S`.
        dropout: The dropout rate on attention weights.
        key: A PRNG key for initialization. If :py:`None`,
            :py:`inox.random.get_rng("init")` is used instead.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int = None,
        hid_features: int = None,
        heads: int = 1,
        bias: bool = True,
        causal: bool = False,
        dropout: Union[float, Array] = 0.0,
        key: Array = None,
    ):
        if key is None:
            keys = get_rng("init").split(4)
        else:
            keys = jax.random.split(key, 4)

        if out_features is None:
            out_features = in_features

        if hid_features is None:
            hid_features = in_features // heads

        self.lin_q = Linear(in_features, hid_features * heads, bias, key=keys[0])
        self.lin_k = Linear(in_features, hid_features * heads, bias, key=keys[1])
        self.lin_v = Linear(in_features, hid_features * heads, bias, key=keys[2])
        self.lin_y = Linear(hid_features * heads, out_features, False, key=keys[3])

        self.heads = heads
        self.causal = causal
        self.dropout = jnp.asarray(dropout)

    def __call__(
        self,
        xq: Array,
        xk: Array = None,
        xv: Array = None,
        mask: Array = None,
        key: Array = None,
    ) -> Array:
        r"""
        Arguments:
            xq: The query tensor :math:`X_q`, with shape :math:`(*, S, C)`.
            xk: The key tensor :math:`X_k`, with shape :math:`(*, T, C)`.
                If :py:`None`, :math:`X_k = X_q`.
            xv: The value tensor :math:`X_v`, with shape :math:`(*, T, C)`.
                If :py:`None`, :math:`X_v = X_k`.
            mask: A boolean attention mask, with shape :math:`(*, N, S, T)`.
                A :py:`False` value indicates that the corresponding attention weight
                is set to :math:`-\infty`.
            key: A PRNG key. If :py:`None`, dropout is not applied.

        Returns:
            The output tensor :math:`Y`, with shape :math:`(*, S, C')`.
        """

        if xk is None:
            xk = xq

        if xv is None:
            xv = xk

        S, T = xq.shape[-2], xk.shape[-2]

        # Project
        q = self.lin_q(xq)
        k = self.lin_k(xk)
        v = self.lin_v(xv)

        q, k, v = [rearrange(x, "... L (N H) -> ... L N H", N=self.heads) for x in (q, k, v)]

        # Mask
        if key is not None:
            shape = jnp.broadcast_shapes(
                (*q.shape[:-3], self.heads, S, 1),
                (*k.shape[:-3], self.heads, 1, T),
                () if mask is None else mask.shape,
            )

            keep = jax.random.bernoulli(key, p=1 - self.dropout, shape=shape)

            if mask is None:
                mask = keep
            else:
                mask = jnp.logical_and(mask, keep)

        # Attention
        if mask is None:
            y = jax.numpy.vectorize(
                lambda q, k, v: jax.nn.dot_product_attention(q, k, v, is_causal=self.causal),
                signature="(S,N,H),(T,N,H),(T,N,H)->(S,N,H)",
            )(q, k, v)
        else:
            y = jax.numpy.vectorize(
                lambda q, k, v, mask: jax.nn.dot_product_attention(
                    q, k, v, mask=mask, is_causal=self.causal
                ),
                signature="(S,N,H),(T,N,H),(T,N,H),(N,S,T)->(S,N,H)",
            )(q, k, v, mask)

        y = rearrange(y, "... N H -> ... (N H)")
        y = self.lin_y(y)

        return y
